/*!@file
** @brief   The implementation of the text builder.
** @remarks The builder creates a text to be copied to the clipboard.
** @author  UETA, Shin-ichi <ueta@usdesign.jp>
**
** @copyright (C) 2014-2020 Ueta System Design Studio\n
** This file is licensed under the terms of the MIT License.\n
** See <https://opensource.org/licenses/MIT>.
*/
#pragma once

#include <ATLComTime.h>

#include <propsys.h>
#include <propkey.h>
#include <propvarutil.h>
#include <shobjidl.h>

#pragma comment(lib, "propsys.lib")

/*!
** @brief   The implementation of the text builder.
** @remarks -
*/
class CTextBuilder
{
public:

	/*!
	** @brief   The property string format.
	** @remarks -
	*/
	static PROPDESC_FORMAT_FLAGS s_pdfFlags;

public:

	/*!
	** @brief   Retrieves a formatted property value of the specified file.
	** @param   [in] pszModifier the string that identifies the property.
	** @param   [in] pszPath the path to the target file or directory.
	** @param   [in] bDisplay true to convert data into a string suitable for display.
	** @returns A CString object that contains the formatted value.
	** @throws  CAtlException if something goes wrong.
	** @remarks -
	*/
	static CString GetFileProperty(LPCWSTR pszModifier, LPCWSTR pszPath, bool bDisplay = false)
	{
		CString strFullPath = GetFullPathName(pszPath);

		HRESULT hResult = S_OK;

		CComPtr<IShellItem2> spShellItem2;
		hResult = ::SHCreateItemFromParsingName(strFullPath, NULL, IID_PPV_ARGS(&spShellItem2));
		if (FAILED(hResult))
			AtlThrow(hResult);

		PROPERTYKEY key;
		hResult = ::PSGetPropertyKeyFromName(pszModifier, &key);
		if (FAILED(hResult))
			AtlThrow(hResult);

		if (bDisplay)
		{
			PROPVARIANT var;
			::PropVariantInit(&var);
			hResult = spShellItem2->GetProperty(key, &var);
			if (FAILED(hResult))
				AtlThrow(hResult);

			LPWSTR pszTaskMem = NULL;

			CString strDisplay;
			hResult = ::PSFormatForDisplayAlloc(key, var, s_pdfFlags, &pszTaskMem);
			strDisplay = pszTaskMem;
			::CoTaskMemFree(pszTaskMem), pszTaskMem = NULL;
			ATLVERIFY(SUCCEEDED(::PropVariantClear(&var)));
			if (FAILED(hResult))
				AtlThrow(hResult);

			#ifdef UNICODE // to remove invisible characters such as format characters

			strDisplay.Remove(L'\x200E'); // LEFT-TO-RIGHT MARK (LRM)
			strDisplay.Remove(L'\x200F'); // RIGHT-TO-LEFT MARK (RLM)

			#endif
			return strDisplay;
		}
		else // retrieve the bare string value
		{
			LPWSTR pszTaskMem = NULL;

			CString strValue;
			hResult = spShellItem2->GetString(key, &pszTaskMem);
			strValue = pszTaskMem;
			::CoTaskMemFree(pszTaskMem), pszTaskMem = NULL;
			if (FAILED(hResult))
				AtlThrow(hResult);

			return strValue;
		}
	}

	/*!
	** @brief   Retrieves a formatted timestamp of the specified file.
	** @param   [in] pszModifier the string that identifies the timestamp.
	** @param   [in] pszPath the path to the target file or directory.
	** @returns A CString object that contains the formatted value.
	** @throws  CAtlException if something goes wrong.
	** @remarks -
	*/
	static CString GetFileTime(LPCTSTR pszModifier, LPCTSTR pszPath)
	{
		WIN32_FILE_ATTRIBUTE_DATA data = {};
		if (!::GetFileAttributesEx(pszPath, GetFileExInfoStandard, &data))
			AtlThrowLastWin32();

		COleDateTime varDateTime;

		if (!pszModifier)
			AtlThrow(E_INVALIDARG);
		else if (_tcsicmp(pszModifier, TEXT("atime")) == 0)
			varDateTime = data.ftLastAccessTime;
		else if (_tcsicmp(pszModifier, TEXT("mtime")) == 0)
			varDateTime = data.ftLastWriteTime;
		else if (_tcsicmp(pszModifier, TEXT("ctime")) == 0)
			varDateTime = data.ftCreationTime;
		else
			AtlThrow(TYPE_E_ELEMENTNOTFOUND);

		ATLASSERT(varDateTime.GetStatus() == COleDateTime::valid);
		CString strDateTime = varDateTime.Format();
		ATLASSERT(!strDateTime.IsEmpty());
		return strDateTime;
	}

	/*!
	** @brief   Retrieves the full path and file name of the specified file.
	** @param   [in] pszPath the path to be resolved.
	** @returns A CString object that contains the full path and file name.
	** @throws  CAtlException if something goes wrong.
	** @remarks -
	*/
	static CString GetFullPathName(LPCTSTR pszPath)
	{
		CString strFullPath;
		DWORD cchFullPath = ::GetFullPathName(pszPath, MAX_PATH, strFullPath.GetBuffer(MAX_PATH), NULL);
		strFullPath.ReleaseBuffer();
		if (cchFullPath == 0)
			AtlThrowLastWin32();
		ATLASSERT(cchFullPath < MAX_PATH);
		ATLASSERT(!strFullPath.IsEmpty());
		return strFullPath;
	}

	/*!
	** @brief   Retrieves the component of the specified path.
	** @param   [in] pszModifier the string that identifies the component.
	** @param   [in] pszPath the path to the target file or directory.
	** @returns A CString object that contains the component.
	** @throws  CAtlException if something goes wrong.
	** @remarks -
	*/
	static CString GetPathComponent(LPCTSTR pszModifier, LPCTSTR pszPath)
	{
		CString strFullPath = GetFullPathName(pszPath);

		LPCTSTR pszFileName = ATLPath::FindFileName(strFullPath);
		ATLASSERT(pszFileName);
		LPCTSTR pszExtension = ATLPath::FindExtension(strFullPath);
		ATLASSERT(pszExtension);

		CString strBaseName(pszFileName, int(pszExtension - pszFileName));
		CString strParent = strFullPath;
		ATLPath::RemoveFileSpec(strParent.GetBuffer(MAX_PATH));
		strParent.ReleaseBuffer();

		if (!pszModifier)
			AtlThrow(E_INVALIDARG);
		else if (_tcsicmp(pszModifier, TEXT("fullpath")) == 0)
			return strFullPath;
		else if (_tcsicmp(pszModifier, TEXT("filename")) == 0)
			return CString(pszFileName);
		else if (_tcsicmp(pszModifier, TEXT("extension")) == 0)
			return CString(pszExtension);
		else if (_tcsicmp(pszModifier, TEXT("basename")) == 0)
			return strBaseName;
		else if (_tcsicmp(pszModifier, TEXT("parent")) == 0)
			return strParent;
		else
			AtlThrow(TYPE_E_ELEMENTNOTFOUND);
	}

	/*!
	** @brief   Retrieves a text to be copied to the clipboard.
	** @param   [out] strText a string that receives the text.
	** @param   [in] argc the count of command-line arguments that follow in argv.
	** @param   [in] argv the array of strings representing command-line arguments.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks To get extended error information, call GetLastError.
	*/
	static BOOL GetText(CString& strText, int argc = __argc, TCHAR *argv[] = __targv)
	{
		try
		{
			enum {

				none, // retrieves non-option arguments as plain text.
				text, // retrieves subsequent arguments as plain text.
				path, // retrieves the components of the paths.
				prop, // retrieves the property values of the files.
				stat, // retrieves information about the files.

			} mode = none;

			CAtlList<CString> lines;
			LPCTSTR pszModifier = NULL;
			bool bDisplay = false;

			for (int i = 1, n = argc; i < n; ++i)
			{
				ATLASSERT(argv);
				ATLASSERT(argv[i]);
				const TCHAR* arg = argv[i];
				const TCHAR* exp = NULL;
				size_t span = 0;

				if (((arg[0] != TEXT('-'))
				  && (arg[0] != TEXT('/'))) || mode)
				{
					CString strLine;

					switch (mode)
					{
					case path:
						strLine = GetPathComponent(pszModifier, arg);
						break;
					case prop:
						strLine = GetFileProperty(pszModifier, arg, bDisplay);
						break;
					case stat:
						strLine = GetFileTime(pszModifier, arg);
						break;
					default:
						strLine = arg; // as plain text
						break;
					}
					lines.AddTail(strLine);
				}
				else if (exp = TEXT("text"), span = _tcslen(exp), _tcsicmp(arg + 1, exp) == 0)
				{
					mode = text, pszModifier = NULL;
				}
				else if (exp = TEXT("path:"), span = _tcslen(exp), _tcsnicmp(arg + 1, exp, span) == 0)
				{
					mode = path, pszModifier = arg + 1 + span;
				}
				else if (exp = TEXT("prop:"), span = _tcslen(exp), _tcsnicmp(arg + 1, exp, span) == 0)
				{
					mode = prop, pszModifier = arg + 1 + span;
				}
				else if (exp = TEXT("display"), span = _tcslen(exp), _tcsicmp(arg + 1, exp) == 0)
				{
					bDisplay = true, pszModifier = NULL;
				}
				else if (exp = TEXT("date:"), span = _tcslen(exp), _tcsnicmp(arg + 1, exp, span) == 0)
				{
					bDisplay = true, pszModifier = arg + 1 + span;

					SetDateFormatForDisplay(pszModifier);
				}
				else if (exp = TEXT("time:"), span = _tcslen(exp), _tcsnicmp(arg + 1, exp, span) == 0)
				{
					bDisplay = true, pszModifier = arg + 1 + span;

					SetTimeFormatForDisplay(pszModifier);
				}
				else if (exp = TEXT("stat:"), span = _tcslen(exp), _tcsnicmp(arg + 1, exp, span) == 0)
				{
					mode = stat, pszModifier = arg + 1 + span;
				}
				else // unknown option
				{
					AtlThrow(E_INVALIDARG);
				}
			}
			if (lines.GetCount() == 1) // single-line text
			{
				strText = lines.GetHead();
			}
			else // multiline text, or empty
			{
				strText.Empty();

				for (POSITION pos = lines.GetHeadPosition(); pos; )
				{
					strText += lines.GetNext(pos) + TEXT("\r\n");
				}
			}
		}
		catch (const CAtlException& cause)
		{
			::SetLastError(DWORD(cause));

			return FALSE;
		}
		return TRUE;
	}

	/*!
	** @brief   Specifies the format to apply to the date.
	** @param   [in] pszModifier the string that identifies the format.
	** @returns No return value.
	** @throws  CAtlException if something goes wrong.
	** @remarks -
	*/
	static void SetDateFormatForDisplay(LPCTSTR pszModifier)
	{
		PROPDESC_FORMAT_FLAGS mask = PDFF_SHORTDATE | PDFF_LONGDATE | PDFF_HIDEDATE;

		if (!pszModifier)
			AtlThrow(E_INVALIDARG);
		else if (_tcsicmp(pszModifier, TEXT("short")) == 0)
			(s_pdfFlags &= ~mask) |= PDFF_SHORTDATE;
		else if (_tcsicmp(pszModifier, TEXT("long")) == 0)
			(s_pdfFlags &= ~mask) |= PDFF_LONGDATE;
		else if (_tcsicmp(pszModifier, TEXT("hide")) == 0)
			(s_pdfFlags &= ~mask) |= PDFF_HIDEDATE;
		else
			AtlThrow(E_INVALIDARG);
	}

	/*!
	** @brief   Specifies the format to apply to the time.
	** @param   [in] pszModifier the string that identifies the format.
	** @returns No return value.
	** @throws  CAtlException if something goes wrong.
	** @remarks -
	*/
	static void SetTimeFormatForDisplay(LPCTSTR pszModifier)
	{
		PROPDESC_FORMAT_FLAGS mask = PDFF_SHORTTIME | PDFF_LONGTIME | PDFF_HIDETIME;

		if (!pszModifier)
			AtlThrow(E_INVALIDARG);
		else if (_tcsicmp(pszModifier, TEXT("short")) == 0)
			(s_pdfFlags &= ~mask) |= PDFF_SHORTTIME;
		else if (_tcsicmp(pszModifier, TEXT("long")) == 0)
			(s_pdfFlags &= ~mask) |= PDFF_LONGTIME;
		else if (_tcsicmp(pszModifier, TEXT("hide")) == 0)
			(s_pdfFlags &= ~mask) |= PDFF_HIDETIME;
		else
			AtlThrow(E_INVALIDARG);
	}
};

PROPDESC_FORMAT_FLAGS CTextBuilder::s_pdfFlags = PDFF_DEFAULT;
