/*!@file
** @brief   The customization of the Date and Time Picker control.
** @remarks Both date and time are displayed in a single control.
** @author  UETA, Shin-ichi <ueta@usdesign.jp>
**
** @copyright (C) 2014-2020 Ueta System Design Studio\n
** This file is licensed under the terms of the MIT License.\n
** See <https://opensource.org/licenses/MIT>.
*/
#pragma once

#include <wininet.h>

#pragma comment(lib, "wininet.lib")

/*!
** @brief   The client side for the customized Date and Time Picker control.
** @remarks -
*/
class CDateTimePickerCtrlEx : public CDateTimePickerCtrl
{
public:

	/*!
	** @brief   Constructs a new CDateTimePickerCtrlEx object with a handle to a control.
	** @param   [in] hWnd a handle to a Date and Time Picker control, or NULL for default.
	** @remarks -
	*/
	CDateTimePickerCtrlEx(HWND hWnd = NULL) : CDateTimePickerCtrl(hWnd)
	{
		;
	}

	/*!
	** @brief   Assigns a handle to a Date and Time Picker control.
	** @param   [in] hWnd a handle to a Date and Time Picker control.
	** @returns By convention, a reference to this object.
	** @remarks -
	*/
	CDateTimePickerCtrlEx& operator =(HWND hWnd)
	{
		CDateTimePickerCtrl::operator =(hWnd);

		return *this;
	}

	/*!
	** @brief   Places the currently selected time in a FILETIME structure based on UTC.
	** @param   [out] lpFileTime a pointer to a FILETIME structure to receive the time.
	** @returns GDT_VALID if successful, GDT_NONE if no date, or GDT_ERROR otherwise.
	** @remarks -
	*/
	DWORD GetFileTime(FILETIME* lpFileTime) const
	{
		#if 1 // daylight saving time aware
		{
			SYSTEMTIME stLocalTime;

			DWORD dwResult = GetSystemTime(&stLocalTime);

			if (dwResult == DWORD(GDT_VALID))
			{
				stLocalTime.wMilliseconds = 0;

				SYSTEMTIME stUniversalTime;

				if (::TzSpecificLocalTimeToSystemTime(NULL, &stLocalTime, &stUniversalTime))
				{
					if (::SystemTimeToFileTime(&stUniversalTime, lpFileTime))
					{
						return DWORD(GDT_VALID);
					}
				}
				return DWORD(GDT_ERROR);
			}
			return dwResult;
		}
		#else
		{
			SYSTEMTIME stLocalTime;

			DWORD dwResult = GetSystemTime(&stLocalTime);

			if (dwResult == DWORD(GDT_VALID))
			{
				stLocalTime.wMilliseconds = 0;

				FILETIME ftLocalFileTime;

				if (::SystemTimeToFileTime(&stLocalTime, &ftLocalFileTime))
				{
					if (::LocalFileTimeToFileTime(&ftLocalFileTime, lpFileTime))
					{
						return DWORD(GDT_VALID);
					}
				}
				return DWORD(GDT_ERROR);
			}
			return dwResult;
		}
		#endif
	}

	/*!
	** @brief   Sets the time contained in a FILETIME structure based on UTC.
	** @param   [in] flag GDT_VALID to set the specified time, or GDT_NONE to set no date.
	** @param   [in] lpFileTime a pointer to a FILETIME structure that contains the time.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	BOOL SetFileTime(DWORD flag, const FILETIME* lpFileTime)
	{
		#if 1 // daylight saving time aware
		{
			SYSTEMTIME stUniversalTime;

			if (::FileTimeToSystemTime(lpFileTime, &stUniversalTime))
			{
				stUniversalTime.wMilliseconds = 0;

				SYSTEMTIME stLocalTime;

				if (::SystemTimeToTzSpecificLocalTime(NULL, &stUniversalTime, &stLocalTime))
				{
					BOOL bResult = SetSystemTime(flag, &stLocalTime);

					return bResult;
				}
			}
			return FALSE;
		}
		#else
		{
			FILETIME ftLocalFileTime;

			if (::FileTimeToLocalFileTime(lpFileTime, &ftLocalFileTime))
			{
				SYSTEMTIME stLocalTime;

				if (::FileTimeToSystemTime(&ftLocalFileTime, &stLocalTime))
				{
					stLocalTime.wMilliseconds = 0;

					BOOL bResult = SetSystemTime(flag, &stLocalTime);

					return bResult;
				}
			}
			return FALSE;
		}
		#endif
	}
};

/*!
** @brief   The implementation of the customized Date and Time Picker control.
** @remarks -
*/
class CDateTimePickerCtrlExImpl : public CWindowImpl<CDateTimePickerCtrlExImpl, CDateTimePickerCtrlEx>
{
public:

	/*!
	** @brief   Initializes a new CDateTimePickerCtrlExImpl object.
	** @remarks -
	*/
	CDateTimePickerCtrlExImpl()
	{
		;
	}

	/*!
	** @brief   Retrieves the standard styles for the customized Date and Time Picker control.
	** @param   [in] dwStyle standard styles used for creation of a window.
	** @returns The default standard styles if dwStyle is zero, or dwStyle otherwise.
	** @remarks -
	*/
	static DWORD GetWndStyle(DWORD dwStyle)
	{
		dwStyle = CWindowImpl<CDateTimePickerCtrlExImpl, CDateTimePickerCtrlEx>::GetWndStyle(dwStyle);

		dwStyle |= DTS_RIGHTALIGN;
		dwStyle |= DTS_APPCANPARSE;
		dwStyle |= DTS_SHORTDATECENTURYFORMAT;

		return dwStyle;
	}

	/*!
	** @brief   Retrieves the extended styles for the customized Date and Time Picker control.
	** @param   [in] dwExStyle extended styles used for creation of a window.
	** @returns The default extended styles if dwExStyle is zero, or dwExStyle otherwise.
	** @remarks -
	*/
	static DWORD GetWndExStyle(DWORD dwExStyle)
	{
		dwExStyle = CWindowImpl<CDateTimePickerCtrlExImpl, CDateTimePickerCtrlEx>::GetWndExStyle(dwExStyle);

		return dwExStyle;
	}

	/*!
	** @brief   Subclasses a Date and Time Picker control and attaches it to the object.
	** @param   [in] hWnd a handle to the Date and Time Picker control to be subclassed.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	BOOL SubclassWindow(HWND hWnd)
	{
		if (!CWindowImpl<CDateTimePickerCtrlExImpl, CDateTimePickerCtrlEx>::SubclassWindow(hWnd))
			return FALSE;

		ATLVERIFY(SetFormat(GetDefaultFormat()));

		return TRUE;
	}

	/*!
	** @brief   Retrieves information about a locale specified by identifier.
	** @param   [in] Locale Locale identifier for which to retrieve information.
	** @param   [in] LCType the locale information to retrieve.
	** @param   [out] LCData a string that receives the requested locale information.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	static BOOL GetLocaleInfo(LCID Locale, LCTYPE LCType, CString& LCData)
	{
		int cchData = ::GetLocaleInfo(Locale, LCType, NULL, 0);

		if (cchData > 0) // including the terminating null character
		{
			::GetLocaleInfo(Locale, LCType, LCData.GetBuffer(cchData), cchData);
			LCData.ReleaseBuffer();
			ATLASSERT(LCData.GetLength() + 1 == cchData);

			return TRUE;
		}
		return FALSE;
	}

	/*!
	** @brief   Retrieves the default format string.
	** @returns The default format string.
	** @remarks -
	*/
	static CString GetDefaultFormat()
	{
		static CString cache;

		if (cache.IsEmpty())
		{
			CString strShortDate;
			ATLVERIFY(GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_SSHORTDATE, strShortDate));
			CString strTimeFormat;
			ATLVERIFY(GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_STIMEFORMAT, strTimeFormat));

			cache = strShortDate + TEXT(' ') + strTimeFormat;
		}
		return cache;
	}

	BEGIN_MSG_MAP(CDateTimePickerCtrlEx)
		MSG_WM_CREATE(OnCreate)
		MSG_WM_SETTEXT(OnSetText)
		REFLECTED_NOTIFY_CODE_HANDLER_EX(DTN_USERSTRING, OnUserString)
	END_MSG_MAP()

	/*!
	** @brief   Called as a part of creation of the window before the window becomes visible.
	** @param   [in] lpCreateStruct a pointer to information about the window being created.
	** @returns Zero to continue creation of the window, or -1 to destroy the window.
	** @remarks -
	*/
	int OnCreate(LPCREATESTRUCT lpCreateStruct)
	{
		LRESULT lResult = DefWindowProc();

		if (lResult == LRESULT(0))
		{
			if (lpCreateStruct
			 && lpCreateStruct->lpszName)
			{
				ATLVERIFY(SetWindowText(lpCreateStruct->lpszName));
			}
			ATLVERIFY(SetFormat(GetDefaultFormat()));
		}
		return int(lResult);
	}

	/*!
	** @brief   Sets the text of a window.
	** @param   [in] lpstrText a pointer to a null-terminated string that is the window text.
	** @returns TRUE if successful, or an error code otherwise.
	** @remarks -
	*/
	int OnSetText(LPCTSTR lpstrText)
	{
		LRESULT lResult = DefWindowProc();

		if (lResult != LRESULT(TRUE)) // FALSE(0), CB_ERR(-1), CB_ERRSPACE(-2), LB_ERRSPACE(-2)
		{
			NMDATETIMESTRING dtString = {{}};
			dtString.nmhdr.hwndFrom = m_hWnd;
			dtString.nmhdr.idFrom = GetDlgCtrlID();
			dtString.nmhdr.code = DTN_USERSTRING;
			dtString.pszUserString = lpstrText;
			if (dtString.pszUserString
			 && dtString.pszUserString[0])
			{
				dtString.dwFlags = GetSystemTime(&dtString.st);
				ATLASSERT(dtString.dwFlags == GDT_VALID);
				OnUserString(&dtString.nmhdr);
			}
			else // current local date and time instead
			{
				::GetLocalTime(&dtString.st);
				dtString.dwFlags = GDT_VALID;
			}
			if (SetSystemTime(dtString.dwFlags, &dtString.st))
			{
				lResult = LRESULT(TRUE);
			}
		}
		return int(lResult);
	}

	/*!
	** @brief   Handles the DTN_USERSTRING notification reflected from the parent window.
	** @param   [in] pnmh a pointer to information about the notification message.
	** @returns The result of the message processing.
	** @remarks -
	*/
	LRESULT OnUserString(LPNMHDR pnmh)
	{
		LPNMDATETIMESTRING lpDTstring = reinterpret_cast<LPNMDATETIMESTRING>(pnmh);

		ATLASSERT(lpDTstring);
		ATLASSERT(lpDTstring->pszUserString);

		CString strUserString = lpDTstring->pszUserString;

		#ifdef UNICODE // to remove invisible characters such as format characters

		strUserString.Remove(L'\x200E'); // LEFT-TO-RIGHT MARK (LRM)
		strUserString.Remove(L'\x200F'); // RIGHT-TO-LEFT MARK (RLM)

		#endif

		DATE date = 0;

		HRESULT hResult = ::VarDateFromStr(strUserString, LOCALE_USER_DEFAULT, 0, &date);

		if (SUCCEEDED(hResult))
		{
			ATLVERIFY(::VariantTimeToSystemTime(date, &lpDTstring->st));

			lpDTstring->dwFlags = GDT_VALID;
		}
		else if (hResult == DISP_E_TYPEMISMATCH)
		{
			SYSTEMTIME stUniversalTime;

			if (::InternetTimeToSystemTime(strUserString, &stUniversalTime, 0)) // as HTTP-date
			{
				if (::SystemTimeToTzSpecificLocalTime(NULL, &stUniversalTime, &lpDTstring->st))
				{
					lpDTstring->dwFlags = GDT_VALID;
				}
			}
		}
		return LRESULT(0);
	}
};
