/*!@file
** @brief   The metadata list based on the List-View control.
** @remarks The metadata of files is provided by the Windows Property System.
** @author  UETA, Shin-ichi <ueta@usdesign.jp>
**
** @copyright (C) 2014-2020 Ueta System Design Studio\n
** This file is licensed under the terms of the MIT License.\n
** See <https://opensource.org/licenses/MIT>.
*/
#pragma once

#include <propsys.h>
#include <propkey.h>
#include <propvarutil.h>
#include <shobjidl.h>

#pragma comment(lib, "propsys.lib")

/*!
** @brief   The implementation of the metadata list.
** @remarks -
*/
class CMetadataListImpl : public CWindowImpl<CMetadataListImpl, CListViewCtrl>
{
private:

	CContainedWindow m_wndFileTime;
	CContainedWindow m_wndDuration;

	struct CProperty
	{
		PROPERTYKEY propkey;
		VARTYPE vartype;

		CString strCanonicalName;
		CString strDisplayName;
		CString strDisplayValue;

		bool bWritable;
		bool bDirty;
	};

	CPath m_path;
	SFGAOF m_sfgaoAttribs;
	CAtlArray<CProperty> m_properties;
	int m_iEditItem, m_iEditSubItem;

public:

	/*!
	** @brief   Initializes a new CMetadataListImpl object.
	** @remarks -
	*/
	CMetadataListImpl() :
			m_wndFileTime(this, 1),
			m_wndDuration(this, 2)
	{
		m_sfgaoAttribs = 0;

		m_iEditItem = -1, m_iEditSubItem = 0;
	}

	/*!
	** @brief   Subclasses a List-View control and attaches it to the object.
	** @param   [in] hWnd a handle to the List-View control to be subclassed.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	BOOL SubclassWindow(HWND hWnd)
	{
		if (!CWindowImpl<CMetadataListImpl, CListViewCtrl>::SubclassWindow(hWnd))
			return FALSE;

		CPropertyPageWindow wndPropertyPage = GetParent();

		ATLVERIFY(m_wndFileTime.SubclassWindow(wndPropertyPage.GetDlgItem(IDC_FILETIME)));
		m_wndFileTime.SetParent(m_hWnd);
		ATLVERIFY(m_wndDuration.SubclassWindow(wndPropertyPage.GetDlgItem(IDC_DURATION)));
		m_wndDuration.SetParent(m_hWnd);

		CRect rcClient;
		ATLVERIFY(GetClientRect(rcClient));
		int cxVScroll = ::GetSystemMetrics(SM_CXVSCROLL);
		SetExtendedListViewStyle(LVS_EX_FULLROWSELECT);

		LVGROUP group = {sizeof(LVGROUP), LVGF_HEADER | LVGF_GROUPID | LVGF_STATE | LVGF_ALIGN};
		CString strGroupHeader;
		ATLVERIFY(strGroupHeader.LoadString(IDS_GROUP_METADATA));
		group.pszHeader = const_cast<LPWSTR>(static_cast<LPCWSTR>(strGroupHeader));
		group.iGroupId = 0;
		group.state = LVGS_NORMAL;
		group.uAlign = LVGA_HEADER_LEFT;
		ATLVERIFY(0 == AddGroup(&group));
		ATLVERIFY(1 == EnableGroupView(TRUE));

		for (int iCol = 0; iCol < 2; ++iCol)
		{
			CString strColumnHeader;
			ATLVERIFY(strColumnHeader.LoadString(IDS_DETAILS_HEADER + iCol));
			ATLVERIFY(iCol == InsertColumn(iCol, strColumnHeader, LVCFMT_LEFT));
		}
		for (size_t iProp = 0, nProps = m_properties.GetCount(); iProp < nProps; ++iProp)
		{
			CProperty& property = m_properties[iProp];

			int iItem = int(iProp);

			ATLVERIFY(iItem == AddItem(iItem, 0, property.strDisplayName));
			ATLVERIFY(iItem == AddItem(iItem, 1, property.strDisplayValue));

			LVITEM item = {LVIF_PARAM | LVIF_GROUPID};
			item.iItem = iItem;
			item.lParam = LPARAM(&property);
			item.iGroupId = group.iGroupId;
			ATLVERIFY(SetItem(&item));
		}
		ATLVERIFY(SetColumnWidth(0, LVSCW_AUTOSIZE_USEHEADER));
		int cxColumn = GetColumnWidth(0);
		cxColumn = rcClient.Width() - cxVScroll * 2 - cxColumn;
		ATLVERIFY(SetColumnWidth(1, cxColumn));

		return TRUE;
	}

	/*!
	** @brief   Begins in-place editing of the specified subitem's text.
	** @param   [in] iItem the index of the subitem's parent item.
	** @param   [in] iSubItem the one-based index of the subitem.
	** @returns No return value.
	** @remarks -
	*/
	void EditSubItem(int iItem, int iSubItem)
	{
		if (m_sfgaoAttribs & SFGAO_READONLY)
			return;

		CProperty* pProperty = reinterpret_cast<CProperty*>(GetItemData(iItem));
		ATLASSERT(pProperty);
		if (!pProperty->bWritable)
			return;

		CString strText;
		GetItemText(iItem, iSubItem, strText);
		ATLVERIFY(EnsureVisible(iItem, FALSE));

		CRect rcClient, rcSelect, rcLabel, rcBounds, rcCtrl;
		ATLVERIFY(GetClientRect(rcClient));
		ATLVERIFY(GetItemRect(iItem, rcSelect, LVIR_SELECTBOUNDS));
		ATLVERIFY(GetItemRect(iItem, rcLabel, LVIR_LABEL));
		ATLVERIFY(GetSubItemRect(iItem, iSubItem, LVIR_BOUNDS, rcBounds));
		int cxMargine = rcLabel.left - rcSelect.left;
		int cyBorder = 1;
		rcCtrl = rcBounds;
		rcCtrl.DeflateRect(cxMargine, 0);
		if (rcCtrl.right < rcCtrl.left)
			rcCtrl.right = rcCtrl.left;
		rcCtrl.IntersectRect(rcClient, rcCtrl);
		rcCtrl.InflateRect(0, cyBorder);

		CWindow wndCtrl;
		switch (iSubItem)
		{
		case 1:
			switch (pProperty->vartype)
			{
			case VT_FILETIME:
				wndCtrl = m_wndFileTime;
				break;
			case VT_UI8:
				wndCtrl = m_wndDuration;
				break;
			default:
				break;
			}
			break;
		default:
			return;
		}
		ATLASSERT(wndCtrl);
		ATLVERIFY(wndCtrl.SetWindowText(strText));
		ATLVERIFY(wndCtrl.SetWindowPos(HWND_TOP, rcCtrl, SWP_SHOWWINDOW));
		CPropertyPageWindow wndPropertyPage = GetParent();
		wndPropertyPage.GotoDlgCtrl(wndCtrl);

		m_iEditItem = iItem, m_iEditSubItem = iSubItem;
	}

	/*!
	** @brief   Ends in-place editing of the subitem's text.
	** @returns No return value.
	** @remarks -
	*/
	void EndEditSubItem()
	{
		CPropertyPageWindow wndPropertyPage = GetParent();
		if (IsChild(GetFocus()))
			wndPropertyPage.GotoDlgCtrl(m_hWnd);
	}

	/*!
	** @brief   Ends in-place editing and applies the specified text to the subitem.
	** @param   [in] lpszText a null-terminated string containing the subitem's text.
	** @returns No return value.
	** @remarks -
	*/
	void EndEditSubItem(LPCTSTR lpszText)
	{
		int iItem = m_iEditItem, iSubItem = m_iEditSubItem;
		m_iEditItem = -1, m_iEditSubItem = 0;
		EndEditSubItem();
		if (iItem < 0)
			return;

		CString strExpired;
		GetItemText(iItem, iSubItem, strExpired);
		if (strExpired == lpszText)
			return;

		ATLVERIFY(SetItemText(iItem, iSubItem, lpszText));
		CPropertyPageWindow wndPropertyPage = GetParent();
		wndPropertyPage.SetModified();

		CProperty* pProperty = reinterpret_cast<CProperty*>(GetItemData(iItem));
		ATLASSERT(pProperty);
		pProperty->bDirty = true;
	}

	/*!
	** @brief   Cancels in-place editing and discards any changes.
	** @returns No return value.
	** @remarks -
	*/
	void CancelEditSubItem()
	{
		m_iEditItem = -1, m_iEditSubItem = 0;
		EndEditSubItem();
	}

	/*!
	** @brief   Checks if any property has changed since it was last saved.
	** @returns true if any property has changed, or false otherwise.
	** @remarks -
	*/
	bool IsDirty() const
	{
		for (size_t iProp = 0, nProps = m_properties.GetCount(); iProp < nProps; ++iProp)
		{
			if (m_properties[iProp].bDirty)
				return true;
		}
		return false;
	}

	/*!
	** @brief   Binds this metadata list to a property store for an item.
	** @param   [in] pszPath a null-terminated string that specifies the item path.
	** @returns S_OK if successful, or an HRESULT error code otherwise.
	** @remarks -
	*/
	HRESULT Bind(LPCWSTR pszPath)
	{
		m_path = pszPath;

		return Load();
	}

	/*!
	** @brief   Reads the properties from the property store.
	** @returns S_OK if successful, or an HRESULT error code otherwise.
	** @remarks -
	*/
	HRESULT Load()
	{
		HRESULT hResult = S_OK;

		GETPROPERTYSTOREFLAGS flags = GPS_DEFAULT | GPS_HANDLERPROPERTIESONLY;

		CComPtr<IShellItem2> spShellItem2;
		hResult = ::SHCreateItemFromParsingName(m_path, NULL, IID_PPV_ARGS(&spShellItem2));
		CComPtr<IPropertyStore> spPropertyStore;
		if (SUCCEEDED(hResult))
			hResult = spShellItem2->GetPropertyStore(flags, IID_PPV_ARGS(&spPropertyStore));
		if (SUCCEEDED(hResult))
		{
			if (m_properties.IsEmpty())
			{
				hResult = spShellItem2->GetAttributes(SFGAO_STORAGECAPMASK, &m_sfgaoAttribs);
				ATLASSERT(SUCCEEDED(hResult));
				if (FAILED(hResult))
					goto OnFailed;

				CComQIPtr<IPropertyStoreCapabilities> spPropertyStoreCapabilities = spPropertyStore;

				REFPROPERTYKEY keyType = PKEY_PropList_FullDetails;
				CComPtr<IPropertyDescriptionList> spPropertyDescriptionList;
				hResult = spShellItem2->GetPropertyDescriptionList(keyType, IID_PPV_ARGS(&spPropertyDescriptionList));
				ATLASSERT(SUCCEEDED(hResult));
				if (FAILED(hResult))
					goto OnFailed;

				UINT cElems = 0;
				hResult = spPropertyDescriptionList->GetCount(&cElems);
				ATLASSERT(SUCCEEDED(hResult));
				if (FAILED(hResult))
					goto OnFailed;

				for (UINT iElem = 0; iElem < cElems; ++iElem)
				{
					CComPtr<IPropertyDescription> spPropertyDescription;
					hResult = spPropertyDescriptionList->GetAt(iElem, IID_PPV_ARGS(&spPropertyDescription));
					ATLASSERT(SUCCEEDED(hResult));
					if (FAILED(hResult))
						goto OnFailed;

					CProperty property;
					property.bWritable = true;
					property.bDirty = false;

					hResult = spPropertyDescription->GetPropertyKey(&property.propkey);
					ATLASSERT(SUCCEEDED(hResult));
					if (FAILED(hResult))
						goto OnFailed;
					hResult = spPropertyDescription->GetPropertyType(&property.vartype);
					ATLASSERT(SUCCEEDED(hResult));
					if (FAILED(hResult))
						goto OnFailed;

					CString& strCanonicalName = property.strCanonicalName;
					LPWSTR pszCanonicalNamee = NULL;
					hResult = spPropertyDescription->GetCanonicalName(&pszCanonicalNamee);
					ATLASSERT(SUCCEEDED(hResult));
					strCanonicalName = pszCanonicalNamee;
					::CoTaskMemFree(pszCanonicalNamee);
					if (FAILED(hResult))
						goto OnFailed;

					CString& strDisplayName = property.strDisplayName;
					LPWSTR pszDisplayName = NULL;
					hResult = spPropertyDescription->GetDisplayName(&pszDisplayName);
					ATLASSERT(SUCCEEDED(hResult));
					strDisplayName = pszDisplayName;
					::CoTaskMemFree(pszDisplayName);
					if (FAILED(hResult))
						goto OnFailed;

					if (spPropertyStoreCapabilities)
					{
						hResult = spPropertyStoreCapabilities->IsPropertyWritable(property.propkey);
						ATLASSERT(SUCCEEDED(hResult));
						if (hResult == S_FALSE)
							property.bWritable = false;

						hResult = S_OK;
					}
					switch (property.vartype)
					{
					case VT_FILETIME:
						if (IsEqualPropertyKey(property.propkey, PKEY_DateAccessed))
							continue;
						if (IsEqualPropertyKey(property.propkey, PKEY_DateCreated))
							continue;
						if (IsEqualPropertyKey(property.propkey, PKEY_DateModified))
							continue;
						break;
					case VT_UI8:
						if (IsEqualPropertyKey(property.propkey, PKEY_Document_TotalEditingTime))
							break;
						continue;
					default:
						continue;
					}
					_ATLTRY
					{
						m_properties.Add(property);
					}
					_ATLCATCH(ex)
					{
						hResult = ex;
						goto OnFailed;
					}
				}
			}
			if (m_properties.IsEmpty())
			{
				hResult = TYPE_E_ELEMENTNOTFOUND;
			}
			if (FAILED(hResult))
				goto OnFailed;

			for (size_t iProp = 0, nProps = m_properties.GetCount(); iProp < nProps; ++iProp)
			{
				CProperty& property = m_properties[iProp];

				CComPtr<IPropertyDescription> spPropertyDescription;
				hResult = ::PSGetPropertyDescription(property.propkey, IID_PPV_ARGS(&spPropertyDescription));
				ATLASSERT(SUCCEEDED(hResult));
				if (FAILED(hResult))
					goto OnFailed;

				PROPVARIANT var;
				::PropVariantInit(&var);
				hResult = spPropertyStore->GetValue(property.propkey, &var);
				ATLASSERT(SUCCEEDED(hResult));
				if (FAILED(hResult))
					goto OnFailed;

				PROPDESC_FORMAT_FLAGS pdfFlags = PDFF_DEFAULT | PDFF_LONGTIME;

				CString& strDisplayValue = property.strDisplayValue;
				LPWSTR pszDisplayValue = NULL;
				hResult = spPropertyDescription->FormatForDisplay(var, pdfFlags, &pszDisplayValue);
				strDisplayValue = pszDisplayValue;
				::CoTaskMemFree(pszDisplayValue);
				ATLASSERT(SUCCEEDED(hResult));
				ATLVERIFY(SUCCEEDED(::PropVariantClear(&var)));
				if (FAILED(hResult))
					goto OnFailed;

				#ifdef UNICODE // to remove invisible characters such as format characters

				strDisplayValue.Remove(L'\x200E'); // LEFT-TO-RIGHT MARK (LRM)
				strDisplayValue.Remove(L'\x200F'); // RIGHT-TO-LEFT MARK (RLM)

				#endif

				if (m_hWnd)
				{
					int iItem = int(iProp);

					ATLVERIFY(SetItemText(iItem, 1, strDisplayValue));
				}
				property.bDirty = false;
			}
		}
	OnFailed:
		return hResult;
	}

	/*!
	** @brief   Writes the modified properties back to the property store.
	** @returns S_OK if successful, or an HRESULT error code otherwise.
	** @remarks -
	*/
	HRESULT Save()
	{
		HRESULT hResult = S_OK;
		if (!IsDirty())
			return hResult;

		GETPROPERTYSTOREFLAGS flags = GPS_READWRITE | GPS_HANDLERPROPERTIESONLY;

		CComPtr<IShellItem2> spShellItem2;
		hResult = ::SHCreateItemFromParsingName(m_path, NULL, IID_PPV_ARGS(&spShellItem2));
		CComPtr<IPropertyStore> spPropertyStore;
		if (SUCCEEDED(hResult))
			hResult = spShellItem2->GetPropertyStore(flags, IID_PPV_ARGS(&spPropertyStore));
		if (SUCCEEDED(hResult))
		{
			for (size_t iProp = 0, nProps = m_properties.GetCount(); iProp < nProps; ++iProp)
			{
				CProperty& property = m_properties[iProp];

				if (!property.bDirty)
					continue;

				CComPtr<IPropertyDescription> spPropertyDescription;
				hResult = ::PSGetPropertyDescription(property.propkey, IID_PPV_ARGS(&spPropertyDescription));
				ATLASSERT(SUCCEEDED(hResult));
				if (FAILED(hResult))
					goto OnFailed;

				CString& strDisplayValue = property.strDisplayValue;
				int iItem = int(iProp);
				GetItemText(iItem, 1, strDisplayValue);
				PROPVARIANT var;
				switch (property.vartype)
				{
				case VT_FILETIME:
					hResult = InitPropFileTimeFromString(strDisplayValue, &var);
					break;
				case VT_UI8:
					hResult = InitPropDurationFromString(strDisplayValue, &var);
					break;
				default:
					hResult = ::InitPropVariantFromString(strDisplayValue, &var);
					break;
				}
				if (SUCCEEDED(hResult))
					hResult = spPropertyDescription->CoerceToCanonicalValue(&var);
				if (SUCCEEDED(hResult))
					hResult = spPropertyStore->SetValue(property.propkey, var);

				ATLASSERT(SUCCEEDED(hResult));
				ATLVERIFY(SUCCEEDED(::PropVariantClear(&var)));
				if (FAILED(hResult))
					goto OnFailed;
			}
			hResult = spPropertyStore->Commit();
		}
	OnFailed:
		return hResult;
	}

	/*!
	** @brief   Initializes a VT_FILETIME propvariant based on a specified string.
	** @param   [in] psz a buffer that contains the source Unicode string.
	** @param   [out] ppropvar an uninitialized PROPVARIANT structure that is initialized.
	** @returns S_OK if successful, or an HRESULT error code otherwise.
	** @remarks -
	*/
	static HRESULT InitPropFileTimeFromString(LPCWSTR psz, PROPVARIANT* ppropvar)
	{
		::PropVariantInit(ppropvar); // to ensure initialization of the propvariant

		if (!psz)
			return E_INVALIDARG;
		if (!psz[0])
			return S_OK; // as VT_EMPTY

		HRESULT hResult = S_OK;

		DATE date;
		SYSTEMTIME st;
		FILETIME ft;

		hResult = ::VarDateFromStr(psz, LOCALE_USER_DEFAULT, 0, &date);

		if (SUCCEEDED(hResult))
		{
			if (::VariantTimeToSystemTime(date, &st))
			{
				if (::TzSpecificLocalTimeToSystemTime(NULL, &st, &st))
				{
					if (st.wYear < 1601)
						return DISP_E_OVERFLOW;
					if (st.wYear > 9999)
						return DISP_E_OVERFLOW;

					if (::SystemTimeToFileTime(&st, &ft))
					{
						hResult = ::InitPropVariantFromFileTime(&ft, ppropvar);
					}
					else
						hResult = AtlHresultFromLastError();
				}
				else
					hResult = AtlHresultFromLastError();
			}
			else
				hResult = AtlHresultFromLastError();
		}
		return hResult;
	}

	/*!
	** @brief   Initializes a VT_UI8 propvariant based on a specified string.
	** @param   [in] psz a buffer that contains the source Unicode string.
	** @param   [out] ppropvar an uninitialized PROPVARIANT structure that is initialized.
	** @returns S_OK if successful, or an HRESULT error code otherwise.
	** @remarks A maximum of up to 512409557:36:10.955.
	*/
	static HRESULT InitPropDurationFromString(LPCWSTR psz, PROPVARIANT* ppropvar)
	{
		::PropVariantInit(ppropvar); // to ensure initialization of the propvariant

		if (!psz)
			return E_INVALIDARG;
		if (!psz[0])
			return S_OK; // as VT_EMPTY

		HRESULT hResult = S_OK;

		unsigned int hours        = 0;
		unsigned int minutes      = 0;
		unsigned int seconds      = 0;
		unsigned int milliseconds = 0;

		int length = int(_tcslen(psz));
		int read = 0, n1 = 0, n2 = 0;

		// hh:mm
		// hh:mm:ss
		// hh:mm:ss.fff

		_stscanf_s(psz, TEXT(" %9u : %2u %n: %2u %n. %n%3u%n %n"), &hours, &minutes, &read, &seconds, &read, &n1, &milliseconds, &n2, &read);

		static const UINT64 maxSeconds = MAXUINT64 / 10000000; // in 100-nanosecond units

		if (read < length)
			return DISP_E_TYPEMISMATCH;
		if (hours > maxSeconds / 60 / 60)
			return DISP_E_OVERFLOW;
		if (minutes > 59)
			return DISP_E_TYPEMISMATCH;
		if (seconds > 59)
			return DISP_E_TYPEMISMATCH;
		if (milliseconds > 999)
			return DISP_E_TYPEMISMATCH;

		if (n2 - n1 == 1)
			milliseconds *= 100; // as hh:mm:ss.f00
		if (n2 - n1 == 2)
			milliseconds *=  10; // as hh:mm:ss.ff0

		UINT64 accSeconds = 0;

		accSeconds += UINT64(hours)   * 60 * 60;
		accSeconds += UINT64(minutes) * 60;
		accSeconds += UINT64(seconds);

		if (accSeconds > maxSeconds)
			return DISP_E_OVERFLOW;
		if (milliseconds > (MAXUINT64 - accSeconds * 10000000) / 10000)
			return DISP_E_OVERFLOW;

		ULONGLONG ullVal = accSeconds * 10000000 + milliseconds * 10000;
		#if 1 // NOTE: Exceeding 32-bit minutes causes overflow or crash in msoshext.dll.
		ullVal = __min(ullVal, UINT64(MAXUINT32) * 60 * 10000000);
		#endif
		hResult = ::InitPropVariantFromUInt64(ullVal, ppropvar);
		ATLASSERT(SUCCEEDED(hResult));
		return hResult;
	}

	/*!
	** @brief   Validate and normalize a string as a date and time.
	** @param   [in, out] strText a string that represents a date and time.
	** @returns true if successful, or false otherwise.
	** @remarks -
	*/
	static bool NormalizeFileTime(CSimpleString& strText)
	{
		HRESULT hResult = S_OK;
		PROPVARIANT var;
		hResult = InitPropFileTimeFromString(strText, &var);
		if (FAILED(hResult))
			return false;
		if (var.vt == VT_EMPTY)
			return true;

		ATLVERIFY(SUCCEEDED(::PropVariantClear(&var)));

		return true; // no need to normalize
	}

	/*!
	** @brief   Validate and normalize a string as a duration.
	** @param   [in, out] strText a string that represents a duration.
	** @returns true if successful, or false otherwise.
	** @remarks -
	*/
	static bool NormalizeDuration(CSimpleString& strText)
	{
		HRESULT hResult = S_OK;
		PROPVARIANT var;
		hResult = InitPropDurationFromString(strText, &var);
		if (FAILED(hResult))
			return false;
		if (var.vt == VT_EMPTY)
			return true;

		PROPDESC_FORMAT_FLAGS pdfFlags = PDFF_DEFAULT | PDFF_LONGTIME;

		LPWSTR pszDisplayValue = NULL;
		hResult = ::PSFormatForDisplayAlloc(PKEY_Document_TotalEditingTime, var, pdfFlags, &pszDisplayValue);
		strText = pszDisplayValue;
		::CoTaskMemFree(pszDisplayValue);
		ATLASSERT(SUCCEEDED(hResult));
		ATLVERIFY(SUCCEEDED(::PropVariantClear(&var)));
		if (FAILED(hResult))
			return false;

		return true;
	}

	BEGIN_MSG_MAP(CMetadataListImpl)
		MSG_WM_KEYDOWN(OnKeyDown)
		MSG_WM_CONTEXTMENU(OnContextMenu)
		REFLECTED_NOTIFY_CODE_HANDLER_EX(NM_DBLCLK, OnDblClk)
		// to finish in-place editing
		REFLECTED_NOTIFY_CODE_HANDLER_EX(LVN_BEGINSCROLL, OnBeginScroll)
		// to finish in-place editing
		NOTIFY_CODE_HANDLER_EX(HDN_BEGINTRACK, OnBeginTrack)
		// Standard clipboard commands
		COMMAND_ID_HANDLER_EX(ID_EDIT_CUT, OnEditCut)
		COMMAND_ID_HANDLER_EX(ID_EDIT_COPY, OnEditCopy)
		COMMAND_ID_HANDLER_EX(ID_EDIT_PASTE, OnEditPaste)
		COMMAND_ID_HANDLER_EX(ID_EDIT_DELETE, OnEditDelete)
		// Commands specific to values of type VT_FILETIME
		COMMAND_RANGE_HANDLER_EX(IDM_FILETIME_FIRST, IDM_FILETIME_LAST, OnSetFileTime)
		// Commands specific to values of type VT_UI8 as duration
		COMMAND_RANGE_HANDLER_EX(IDM_DURATION_FIRST, IDM_DURATION_LAST, OnSetDuration)
	ALT_MSG_MAP(1)
		REFLECTED_NOTIFY_CODE_HANDLER(DTN_USERSTRING, OnUserString1)
		MESSAGE_HANDLER(WM_GETDLGCODE, OnGetDlgCode1)
		REFLECTED_NOTIFY_CODE_HANDLER(NM_KILLFOCUS, OnKillFocus1)
	ALT_MSG_MAP(2)
		MESSAGE_HANDLER(WM_GETDLGCODE, OnGetDlgCode2)
		REFLECTED_COMMAND_CODE_HANDLER(EN_KILLFOCUS, OnKillFocus2)
	END_MSG_MAP()

	/*!
	** @brief   Deletes the currently selected property value and copies the value to the clipboard.
	** @param   [in] uNotifyCode the notification code.
	** @param   [in] nID the identifier of the menu item, control, or accelerator.
	** @param   [in] hWndCtrl a handle to the control.
	** @returns No return value.
	** @remarks -
	*/
	void OnEditCut(UINT uNotifyCode, int nID, HWND hWndCtrl)
	{
		UNREFERENCED_PARAMETER(uNotifyCode);
		UNREFERENCED_PARAMETER(nID);
		UNREFERENCED_PARAMETER(hWndCtrl);

		// NOTE: not implemented
	}

	/*!
	** @brief   Copies currently selected property value to the clipboard.
	** @param   [in] uNotifyCode the notification code.
	** @param   [in] nID the identifier of the menu item, control, or accelerator.
	** @param   [in] hWndCtrl a handle to the control.
	** @returns No return value.
	** @remarks -
	*/
	void OnEditCopy(UINT uNotifyCode, int nID, HWND hWndCtrl)
	{
		UNREFERENCED_PARAMETER(uNotifyCode);
		UNREFERENCED_PARAMETER(nID);
		UNREFERENCED_PARAMETER(hWndCtrl);

		int iItem = GetSelectedIndex(), iSubItem = 1;
		if (iItem < 0)
			return;

		if (OpenClipboard())
		{
			CString strText;
			GetItemText(iItem, iSubItem, strText);
			size_t cchText = wcslen(strText) + 1; // include terminating null character
			if (HGLOBAL hGlobal = ::GlobalAlloc(GMEM_MOVEABLE, sizeof(WCHAR) * cchText))
			{
				LPWSTR pszGlobal = static_cast<LPWSTR>(::GlobalLock(hGlobal));
				ATLASSERT(pszGlobal);
				ATLVERIFY(wcscpy_s(pszGlobal, cchText, strText) == 0);
				ATLVERIFY(::GlobalUnlock(hGlobal) == 0);
				ATLASSERT(::GetLastError() == NO_ERROR);

				ATLVERIFY(::EmptyClipboard());
				// If SetClipboardData succeeds, the system owns the object.
				if (!::SetClipboardData(CF_UNICODETEXT, hGlobal))
					ATLVERIFY((hGlobal = ::GlobalFree(hGlobal)) == NULL);
			}
			ATLVERIFY(::CloseClipboard());
		}
	}

	/*!
	** @brief   Replaces the currently selected property value with the content of the clipboard.
	** @param   [in] uNotifyCode the notification code.
	** @param   [in] nID the identifier of the menu item, control, or accelerator.
	** @param   [in] hWndCtrl a handle to the control.
	** @returns No return value.
	** @remarks -
	*/
	void OnEditPaste(UINT uNotifyCode, int nID, HWND hWndCtrl)
	{
		UNREFERENCED_PARAMETER(uNotifyCode);
		UNREFERENCED_PARAMETER(nID);
		UNREFERENCED_PARAMETER(hWndCtrl);

		// NOTE: not implemented
	}

	/*!
	** @brief   Deletes the currently selected property value.
	** @param   [in] uNotifyCode the notification code.
	** @param   [in] nID the identifier of the menu item, control, or accelerator.
	** @param   [in] hWndCtrl a handle to the control.
	** @returns No return value.
	** @remarks -
	*/
	void OnEditDelete(UINT uNotifyCode, int nID, HWND hWndCtrl)
	{
		UNREFERENCED_PARAMETER(uNotifyCode);
		UNREFERENCED_PARAMETER(nID);
		UNREFERENCED_PARAMETER(hWndCtrl);

		int iItem = GetSelectedIndex(), iSubItem = 1;
		if (iItem < 0)
			return;

		CString strText;
		ATLVERIFY(EnsureVisible(iItem, FALSE));
		m_iEditItem = iItem, m_iEditSubItem = iSubItem;
		EndEditSubItem(strText);
	}

	/*!
	** @brief   Sets the currently selected property value of type VT_FILETIME.
	** @param   [in] uNotifyCode the notification code.
	** @param   [in] nID the identifier of the menu item, control, or accelerator.
	** @param   [in] hWndCtrl a handle to the control.
	** @returns No return value.
	** @remarks -
	*/
	void OnSetFileTime(UINT uNotifyCode, int nID, HWND hWndCtrl)
	{
		UNREFERENCED_PARAMETER(uNotifyCode);
		UNREFERENCED_PARAMETER(hWndCtrl);

		int iItem = GetSelectedIndex(), iSubItem = 1;
		if (iItem < 0)
			return;

		CString strText;
		GetItemText(iItem, iSubItem, strText);
		ATLVERIFY(EnsureVisible(iItem, FALSE));

		ATLVERIFY(m_wndFileTime.SetWindowText(strText));

		CFileTimeEx ft;

		ATLVERIFY(CDateTimePickerCtrlEx(m_wndFileTime).GetFileTime(&ft) == GDT_VALID);

		switch (nID)
		{
		case IDM_FILETIME_CURRENTDATETIME:
			ft = CFileTimeEx::GetCurrentTime().ZeroMillisecond();
			break;
		case IDM_FILETIME_TRUNCTODAY:
			ft = ft.GetTzMidnight();
			break;
		case IDM_FILETIME_TRUNCTOMINUTE:
			ft = ft.ZeroSecond();
			break;
		case IDM_FILETIME_CREATIONTIME:
			if (CFile(m_path).GetFileTime(&ft, NULL, NULL))
				break;
			return;
		case IDM_FILETIME_LASTACCESSTIME:
			if (CFile(m_path).GetFileTime(NULL, &ft, NULL))
				break;
			return;
		case IDM_FILETIME_LASTWRITETIME:
			if (CFile(m_path).GetFileTime(NULL, NULL, &ft))
				break;
			return;
		default:
			return;
		}
		ATLVERIFY(CDateTimePickerCtrlEx(m_wndFileTime).SetFileTime(GDT_VALID, &ft));

		m_wndFileTime.GetWindowText(strText);
		ATLVERIFY(NormalizeFileTime(strText));
		m_iEditItem = iItem, m_iEditSubItem = iSubItem;
		EndEditSubItem(strText);
	}

	/*!
	** @brief   Sets the currently selected property value of type VT_UI8 as a duration.
	** @param   [in] uNotifyCode the notification code.
	** @param   [in] nID the identifier of the menu item, control, or accelerator.
	** @param   [in] hWndCtrl a handle to the control.
	** @returns No return value.
	** @remarks -
	*/
	void OnSetDuration(UINT uNotifyCode, int nID, HWND hWndCtrl)
	{
		UNREFERENCED_PARAMETER(uNotifyCode);
		UNREFERENCED_PARAMETER(hWndCtrl);

		int iItem = GetSelectedIndex(), iSubItem = 1;
		if (iItem < 0)
			return;

		CString strText;
		GetItemText(iItem, iSubItem, strText);
		ATLVERIFY(EnsureVisible(iItem, FALSE));

		ATLVERIFY(m_wndDuration.SetWindowText(strText));

		switch (nID)
		{
		case IDM_DURATION_ZERO:
			ATLVERIFY(m_wndDuration.SetWindowText(TEXT("00:00:00")));
			break;
		default:
			return;
		}
		m_wndDuration.GetWindowText(strText);
		ATLVERIFY(NormalizeDuration(strText));
		m_iEditItem = iItem, m_iEditSubItem = iSubItem;
		EndEditSubItem(strText);
	}

	/*!
	** @brief   Called when a nonsystem key is pressed.
	** @param   [in] nChar the virtual key code of the given key.
	** @param   [in] nRepCnt the number of times the keystroke is autorepeated.
	** @param   [in] nFlags additional information about the keystroke.
	** @returns No return value.
	** @remarks -
	*/
	void OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags)
	{
		UNREFERENCED_PARAMETER(nRepCnt);
		UNREFERENCED_PARAMETER(nFlags);

		SetMsgHandled(FALSE);

		int iItem = GetSelectedIndex(), iSubItem = 1;
		if (iItem < 0)
			return;

		switch (nChar)
		{
		case VK_F2:
			EditSubItem(iItem, iSubItem); // as label editing
			SetMsgHandled(TRUE);
			break;
		default:
			break;
		}
	}

	/*!
	** @brief   Called when the user has clicked the right mouse button.
	** @param   [in] hWnd a handle to the window in which the user right-clicked the mouse.
	** @param   [in] point the position of the cursor, in screen coordinates.
	** @returns No return value.
	** @remarks -
	*/
	void OnContextMenu(HWND hWnd, CPoint point)
	{
		if (IsChild(hWnd))
			return;
		int iItem = GetSelectedIndex(), iSubItem = 1;
		if (iItem < 0)
			return;

		UINT nMenuID;

		CProperty* pProperty = reinterpret_cast<CProperty*>(GetItemData(iItem));
		ATLASSERT(pProperty);
		switch (pProperty->vartype)
		{
		case VT_FILETIME:
			nMenuID = IDR_FILETIME;
			break;
		case VT_UI8:
			nMenuID = IDR_DURATION;
			break;
		default:
			return;
		}
		CMenu menu;
		ATLVERIFY(menu.LoadMenu(nMenuID));
		ATLASSERT(menu.GetMenuItemCount() == 1);
		CMenuHandle menuPopup = menu.GetSubMenu(0);
		ATLASSERT(menuPopup);

		bool bReadOnly = false;

		if (m_sfgaoAttribs & SFGAO_READONLY)
			bReadOnly = true;
		if (!pProperty->bWritable)
			bReadOnly = true;
		if (bReadOnly)
		{
			for (int i = 0, n = menuPopup.GetMenuItemCount(); i < n; ++i)
			{
				int nID = menuPopup.GetMenuItemID(i);
				ATLASSERT(nID != -1);
				switch (nID)
				{
				case 0:
					continue;
				case ID_EDIT_COPY:
					continue;
				default:
					break;
				}
				menuPopup.EnableMenuItem(nID, MF_GRAYED);
			}
		}
		CString strText;
		GetItemText(iItem, iSubItem, strText);
		if (strText.IsEmpty())
		{
			menuPopup.EnableMenuItem(IDM_FILETIME_TRUNCTODAY,    MF_GRAYED);
			menuPopup.EnableMenuItem(IDM_FILETIME_TRUNCTOMINUTE, MF_GRAYED);
		}
		if ((point.x == -1)
		 && (point.y == -1)) // generated by Applications key or SHIFT+F10
		{
			CRect rcSelect;
			ATLVERIFY(GetItemRect(iItem, rcSelect, LVIR_SELECTBOUNDS));
			ATLVERIFY(ClientToScreen(rcSelect));
			point = rcSelect.BottomRight();
		}
		if (!menuPopup.TrackPopupMenu(0, point.x, point.y, m_hWnd))
			ATLASSERT(::GetLastError() == ERROR_POPUP_ALREADY_ACTIVE);
	}

	/*!
	** @brief   Handles the NM_DBLCLK notification reflected from the parent window.
	** @param   [in] pnmh a pointer to information about the notification message.
	** @returns The result of the message processing.
	** @remarks -
	*/
	LRESULT OnDblClk(LPNMHDR pnmh)
	{
		LPNMITEMACTIVATE pnmItemActivate = reinterpret_cast<LPNMITEMACTIVATE>(pnmh);
		ATLASSERT(pnmItemActivate);
		switch (pnmItemActivate->iSubItem)
		{
		case 1:
			EditSubItem(pnmItemActivate->iItem, pnmItemActivate->iSubItem);
			break;
		default:
			break;
		}
		return DefWindowProc();
	}

	/*!
	** @brief   Handles the LVN_BEGINSCROLL notification reflected from the parent window.
	** @param   [in] pnmh a pointer to information about the notification message.
	** @returns The result of the message processing.
	** @remarks -
	*/
	LRESULT OnBeginScroll(LPNMHDR pnmh)
	{
		LPNMLVSCROLL pnmLVScroll = reinterpret_cast<LPNMLVSCROLL>(pnmh);
		ATLASSERT(pnmLVScroll);
		UNREFERENCED_PARAMETER(pnmLVScroll);

		EndEditSubItem();

		return DefWindowProc();
	}

	/*!
	** @brief   Handles the HDN_BEGINTRACK notification from the header control.
	** @param   [in] pnmh a pointer to information about the notification message.
	** @returns The result of the message processing.
	** @remarks -
	*/
	LRESULT OnBeginTrack(LPNMHDR pnmh)
	{
		LPNMHEADER pnmHeader = reinterpret_cast<LPNMHEADER>(pnmh);
		ATLASSERT(pnmHeader);
		UNREFERENCED_PARAMETER(pnmHeader);

		EndEditSubItem();

		return DefWindowProc();
	}

	/*!
	** @brief   Handles the DTN_USERSTRING notification reflected from the parent window.
	** @param   [in] nID the identifier of the control sending the message.
	** @param   [in] pnmh a pointer to information about the notification message.
	** @param   [in, out] bHandled set to FALSE if further processing is needed.
	** @returns The result of the message processing.
	** @remarks -
	*/
	LRESULT OnUserString1(int nID, LPNMHDR pnmh, BOOL& bHandled)
	{
		LPNMDATETIMESTRING lpDTstring = reinterpret_cast<LPNMDATETIMESTRING>(pnmh);
		ATLASSERT(lpDTstring);
		UNREFERENCED_PARAMETER(nID);
		UNREFERENCED_PARAMETER(bHandled);

		CString strUserString = lpDTstring->pszUserString;

		if (strUserString.Trim().IsEmpty())
		{
			EndEditSubItem(strUserString);
			lpDTstring->dwFlags = GDT_NONE;
			return LRESULT(0);
		}
		return m_wndFileTime.DefWindowProc();
	}

	/*!
	** @brief   Handles the WM_GETDLGCODE message.
	** @param   [in] uMsg specifies the message.
	** @param   [in] wParam additional message-specific information.
	** @param   [in] lParam additional message-specific information.
	** @param   [in, out] bHandled set to FALSE if further processing is needed.
	** @returns The result of the message processing.
	** @remarks -
	*/
	LRESULT OnGetDlgCode1(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
	{
		UNREFERENCED_PARAMETER(uMsg);
		UNREFERENCED_PARAMETER(lParam);
		UNREFERENCED_PARAMETER(bHandled);

		LRESULT lResult = m_wndFileTime.DefWindowProc();
		switch (wParam)
		{
		case VK_RETURN:
			EndEditSubItem();
			lResult |= DLGC_WANTALLKEYS;
			break;
		case VK_ESCAPE:
			CancelEditSubItem();
			lResult |= DLGC_WANTALLKEYS;
			break;
		default:
			break;
		}
		return lResult;
	}

	/*!
	** @brief   Handles the EN_KILLFOCUS notification reflected from the parent window.
	** @param   [in] nID the identifier of the control sending the message.
	** @param   [in] pnmh a pointer to information about the notification message.
	** @param   [in, out] bHandled set to FALSE if further processing is needed.
	** @returns The result of the message processing.
	** @remarks -
	*/
	LRESULT OnKillFocus1(int nID, LPNMHDR pnmh, BOOL& bHandled)
	{
		UNREFERENCED_PARAMETER(nID);
		UNREFERENCED_PARAMETER(pnmh);
		UNREFERENCED_PARAMETER(bHandled);

		m_wndFileTime.ShowWindow(SW_HIDE);

		#if 1 // NOTE: DTP control returns incomplete input for WM_GETTEXT message.
		SYSTEMTIME st;
		CDateTimePickerCtrlEx(m_wndFileTime).GetSystemTime(&st);
		#endif
		CString strText;
		m_wndFileTime.GetWindowText(strText);
		if (NormalizeFileTime(strText.Trim()))
			EndEditSubItem(strText);
		else
			CancelEditSubItem();

		return m_wndFileTime.DefWindowProc();
	}

	/*!
	** @brief   Handles the WM_GETDLGCODE message.
	** @param   [in] uMsg specifies the message.
	** @param   [in] wParam additional message-specific information.
	** @param   [in] lParam additional message-specific information.
	** @param   [in, out] bHandled set to FALSE if further processing is needed.
	** @returns The result of the message processing.
	** @remarks -
	*/
	LRESULT OnGetDlgCode2(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
	{
		UNREFERENCED_PARAMETER(uMsg);
		UNREFERENCED_PARAMETER(lParam);
		UNREFERENCED_PARAMETER(bHandled);

		LRESULT lResult = m_wndDuration.DefWindowProc();
		switch (wParam)
		{
		case VK_RETURN:
			EndEditSubItem();
			lResult |= DLGC_WANTALLKEYS;
			break;
		case VK_ESCAPE:
			CancelEditSubItem();
			lResult |= DLGC_WANTALLKEYS;
			break;
		default:
			break;
		}
		return lResult;
	}

	/*!
	** @brief   Handles the EN_KILLFOCUS notification reflected from the parent window.
	** @param   [in] uNotifyCode the notification code.
	** @param   [in] nID the identifier of the control sending the message.
	** @param   [in] hWndCtrl a handle to the control sending the message.
	** @param   [in, out] bHandled set to FALSE if further processing is needed.
	** @returns The result of the message processing.
	** @remarks -
	*/
	LRESULT OnKillFocus2(UINT uNotifyCode, int nID, HWND hWndCtrl, BOOL& bHandled)
	{
		UNREFERENCED_PARAMETER(uNotifyCode);
		UNREFERENCED_PARAMETER(nID);
		UNREFERENCED_PARAMETER(hWndCtrl);
		UNREFERENCED_PARAMETER(bHandled);

		m_wndDuration.ShowWindow(SW_HIDE);

		CString strText;
		m_wndDuration.GetWindowText(strText);
		if (NormalizeDuration(strText.Trim()))
			EndEditSubItem(strText);
		else
			CancelEditSubItem();

		return m_wndDuration.DefWindowProc();
	}
};
